"""MQCSP: Security Parameters"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, ibmmqc

class CSP(MQOpts):
    """ Construct an MQCSP Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):

        csp_current_version = ibmmqc.__strucversions__.get("csp", 1)

        opts = [['_StrucId', CMQC.MQCSP_STRUC_ID, '4s'],
                ['Version', CMQC.MQCSP_VERSION_1, MQLONG_TYPE],
                ['AuthenticationType', CMQC.MQCSP_AUTH_NONE, MQLONG_TYPE],
                ['_Reserved1', 0, MQLONG_TYPE],
                ['CSPUserId', 0, 'P'],
                ['_CSPUserIdOffset', 0, MQLONG_TYPE],
                ['_CSPUserIdLength', 0, MQLONG_TYPE],
                ['_Reserved2', 0, INTEGER64_TYPE],
                ['CSPPassword', 0, 'P'],
                ['_CSPPasswordOffset', 0, MQLONG_TYPE],
                ['_CSPPasswordLength', 0, MQLONG_TYPE]]

        if csp_current_version >= CMQC.MQCSP_VERSION_2:
            opts += [['_Reserved3', 0, INTEGER64_TYPE],
                     ['InitialKey', 0, 'P'],
                     ['_InitialKeyOffset', 0, MQLONG_TYPE],
                     ['_InitialKeyLength', 0, MQLONG_TYPE]]

        if csp_current_version >= CMQC.MQCSP_VERSION_3:
            opts += [['_Reserved4', 0, INTEGER64_TYPE],
                     ['Token', 0, 'P'],
                     ['_TokenOffset', 0, MQLONG_TYPE],
                     ['_TokenLength', 0, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)



"""MQCIH: CICS Information Header"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class CIH(MQOpts):
    """ Construct an MQCIH Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    The initialisation assumes a CIH Version 2 which has been valid for many years.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQCIH_STRUC_ID, '4s'],
                ['Version', CMQC.MQCIH_VERSION_2, MQLONG_TYPE],
                ['StrucLength', CMQC.MQCIH_LENGTH_2, MQLONG_TYPE],
                ['Encoding', 0, MQLONG_TYPE],
                ['CodedCharSetId', 0, MQLONG_TYPE],
                ['Format', CMQC.MQFMT_NONE, '8s'],
                ['Flags', CMQC.MQCIH_NONE, MQLONG_TYPE],
                ['ReturnCode', CMQC.MQCRC_OK, MQLONG_TYPE],
                ['CompCode', CMQC.MQCC_OK, MQLONG_TYPE],
                ['Reason', CMQC.MQRC_NONE, MQLONG_TYPE],
                ['UOWControl', CMQC.MQCUOWC_ONLY, MQLONG_TYPE],
                ['GetWaitInterval', CMQC.MQCGWI_DEFAULT, MQLONG_TYPE],
                ['LinkType', CMQC.MQCLT_PROGRAM, MQLONG_TYPE],
                ['OutputDataLength', CMQC.MQCODL_AS_INPUT, MQLONG_TYPE],
                ['FacilityKeepTime', 0, MQLONG_TYPE],
                ['ADSDescriptor', CMQC.MQCADSD_NONE, MQLONG_TYPE],
                ['ConversationalTask', CMQC.MQCCT_NO, MQLONG_TYPE],
                ['TaskEndStatus', CMQC.MQCTES_NOSYNC, MQLONG_TYPE],
                ['Facility', CMQC.MQCFAC_NONE, '8s'],
                ['Function', CMQC.MQCFUNC_NONE, '4s'],
                ['AbendCode', ' ', '4s'],
                ['Authenticator', ' ', '8s'],
                ['_Reserved1', ' ', '8s'],
                ['ReplyToFormat', CMQC.MQFMT_NONE, '8s'],
                ['RemoteSysId', ' ', '4s'],
                ['RemoteTransId', ' ', '4s'],
                ['TransactionId', ' ', '4s'],
                ['FacilityLike', ' ', '4s'],
                ['AttentionId', ' ', '4s'],
                ['StartCode', CMQC.MQCSC_NONE, '4s'],
                ['CancelCode', ' ', '4s'],
                ['NextTransactionId', ' ', '4s'],
                ['_Reserved2', ' ', '8s'],
                ['_Reserved3', ' ', '8s'],
                ['CursorPosition', 0, MQLONG_TYPE],
                ['ErrorOffset', 0, MQLONG_TYPE],
                ['InputItem', 0, MQLONG_TYPE],
                ['_Reserved4', 0, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)



"""
Created on 15 Nov 2010

@author: hannes
"""

import os
import unittest
import ibmmq as mq
from ibmmq import CMQC

# pylint: disable=line-too-long,broad-exception-caught

class TestRFH2(unittest.TestCase):
    """This test case tests the RFH2 class and it's methods.
    """

    messages_dir = os.path.join(os.path.dirname(__file__), "messages")

    def setUp(self):
        """ Create a new queue manager (TESTPMQI).
        Must be run as a user that has 'mqm' access.

        """
        with open(os.path.join(self.messages_dir, "single_rfh2.dat"), "rb") as file:
            self.single_rfh2_message = file.read()
        self.single_rfh2_message_not_well_formed = \
            self.single_rfh2_message[0:117] + self.single_rfh2_message[121:]

    def test_parse_rfh2(self):
        """Test that a known correct 3rd party message parses correctly.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message)
            self.assertEqual(len(rfh2.get()), 14,
                 "Number of attributes incorrect.  Should be %i? But is %s"
                     % (14, str(len(rfh2.get()))))
            self.assertEqual(rfh2["_StrucId"], CMQC.MQRFH_STRUC_ID,
                 "StrucId has incorrect value. Should be: %s But is: %s"
                     % (CMQC.MQRFH_STRUC_ID, str(rfh2["_StrucId"])))
            self.assertEqual(rfh2["Version"], CMQC.MQRFH_VERSION_2,
                 "Version has incorrect value. Should be: %i But is: %s"
                     % (CMQC.MQRFH_VERSION_2, str(rfh2["Version"])))
            self.assertEqual(rfh2["StrucLength"], 284,
                 "StrucLength has incorrect value. Should be: %i But is: %s"
                     % (284, str(rfh2["StrucLength"])))
            self.assertEqual(rfh2["Encoding"], 273,
                 "Encoding has incorrect value. Should be: %i But is: %s"
                     % (273, str(rfh2["Encoding"])))
            self.assertEqual(rfh2["CodedCharSetId"], 1208,
                 "CodedCharSetId has incorrect value. Should be: %i But is: %s"
                     % (1208, str(rfh2["CodedCharSetId"])))
            self.assertEqual(rfh2["Format"], CMQC.MQFMT_STRING,
                 "Format has incorrect value. Should be: %s But is: %s"
                     % (CMQC.MQFMT_NONE, str(rfh2["Format"])))
            self.assertEqual(rfh2["Flags"], 0,
                 "Flags has incorrect value. Should be: %i But is: %s"
                     % (0, str(rfh2["Flags"])))
            self.assertEqual(rfh2["NameValueCCSID"], 1208,
                 "NameValueCCSID has incorrect value. Should be: %i But is: %s"
                     % (1208, str(rfh2["NameValueCCSID"])))
            self.assertEqual(rfh2["pscLength"], 152,
                 "pscLength has incorrect value. Should be: %i But is: %s"
                     % (152, str(rfh2["pscLength"])))
            self.assertEqual(rfh2["psc"],
                 b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ",
                 "psc has incorrect value. Should be: %s But is: %s"
                    % ("<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ", ">" + str(rfh2["psc"]) + "<"))
            self.assertEqual(rfh2["testFolderLength"], 56,
                 "testFolderLength has incorrect value. Should be: %i But is: %s"
                     % (56, str(rfh2["testFolderLength"])))
            self.assertEqual(rfh2["testFolder"],
                 b"<testFolder><testVar>testValue</testVar></testFolder>   ",
                 "testFolder has incorrect value. Should be: %s But is: %s"
                     % ("<testFolder><testVar>testValue</testVar></testFolder>   ", str(rfh2["testFolder"])))
            self.assertEqual(rfh2["mcdLength"], 28,
                 "mcdLength has incorrect value. Should be: %i But is: %s"
                     % (28, str(rfh2["mcdLength"])))
            self.assertEqual(rfh2["mcd"],
                 b"<mcd><Msd>xmlnsc</Msd></mcd>",
                 "mcd has incorrect value. Should be: %s But is: %s"
                     % ("<mcd><Msd>xmlnsc</Msd></mcd>", str(rfh2["mcd"])))
        except Exception as e:
            self.fail(e)

    def test_parse_default_rfh2(self):
        """Test parsing of default RFH2 only.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message[0:36])
            self.assertEqual(len(rfh2.get()), 8,
                             "Number of attributes incorrect.  Should be 8? But is %s"
                             % str(len(rfh2.get())))
            self.assertEqual(rfh2["_StrucId"], CMQC.MQRFH_STRUC_ID,
                             "StrucId has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQRFH_STRUC_ID, str(rfh2["_StrucId"])))
            self.assertEqual(rfh2["Version"], CMQC.MQRFH_VERSION_2,
                             "Version has incorrect value. Should be: %i But is: %s"
                             % (CMQC.MQRFH_VERSION_2, str(rfh2["Version"])))
            self.assertEqual(rfh2["StrucLength"], 284,
                             "StrucLength has incorrect value. Should be: %i But is: %s"
                             % (284, str(rfh2["StrucLength"])))
            self.assertEqual(rfh2["Encoding"], 273,
                             "Encoding has incorrect value. Should be: %i But is: %s"
                             % (273, str(rfh2["Encoding"])))
            self.assertEqual(rfh2["CodedCharSetId"], 1208,
                             "CodedCharSetId has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["CodedCharSetId"])))
            self.assertEqual(rfh2["Format"], CMQC.MQFMT_STRING,
                             "Format has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQFMT_NONE, str(rfh2["Format"])))
            self.assertEqual(rfh2["Flags"], 0,
                             "Flags has incorrect value. Should be: %i But is: %s"
                             % (0, str(rfh2["Flags"])))
            self.assertEqual(rfh2["NameValueCCSID"], 1208,
                             "NameValueCCSID has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["NameValueCCSID"])))
        except Exception as e:
            self.fail(e)

    def test_parse_rfh2_with_correct_encoding(self):
        """Parse known correct message with it's correct encoding.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message, 273)
            self.assertEqual(len(rfh2.get()), 14,
                             "Number of attributes incorrect.  Should be 12? But is %s"
                             % str(len(rfh2.get())))
            self.assertEqual(rfh2["_StrucId"], CMQC.MQRFH_STRUC_ID,
                             "StrucId has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQRFH_STRUC_ID, str(rfh2["_StrucId"])))
            self.assertEqual(rfh2["Version"], CMQC.MQRFH_VERSION_2,
                             "Version has incorrect value. Should be: %i But is: %s"
                             % (CMQC.MQRFH_VERSION_2, str(rfh2["Version"])))
            self.assertEqual(rfh2["StrucLength"], 284,
                             "StrucLength has incorrect value. Should be: %i But is: %s"
                             % (284, str(rfh2["StrucLength"])))
            self.assertEqual(rfh2["Encoding"], 273,
                             "Encoding has incorrect value. Should be: %i But is: %s"
                             % (273, str(rfh2["Encoding"])))
            self.assertEqual(rfh2["CodedCharSetId"], 1208,
                             "CodedCharSetId has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["CodedCharSetId"])))
            self.assertEqual(rfh2["Format"], CMQC.MQFMT_STRING,
                             "Format has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQFMT_NONE, str(rfh2["Format"])))
            self.assertEqual(rfh2["Flags"], 0,
                             "Flags has incorrect value. Should be: %i But is: %s"
                             % (0, str(rfh2["Flags"])))
            self.assertEqual(rfh2["NameValueCCSID"], 1208,
                             "NameValueCCSID has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["NameValueCCSID"])))
            self.assertEqual(rfh2["pscLength"], 152,
                             "pscLength has incorrect value. Should be: %i But is: %s"
                             % (152, str(rfh2["pscLength"])))
            self.assertEqual(rfh2["psc"],
                             b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ",
                             "psc has incorrect value. Should be: %s But is: %s"
                             % ("<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ", ">" + str(rfh2["psc"]) + "<"))
            self.assertEqual(rfh2["testFolderLength"], 56,
                             "testFolderLength has incorrect value. Should be: %i But is: %s"
                             % (56, str(rfh2["testFolderLength"])))
            self.assertEqual(rfh2["testFolder"], b"<testFolder><testVar>testValue</testVar></testFolder>   ",
                             "testFolder has incorrect value. Should be: %s But is: %s"
                             % ("<testFolder><testVar>testValue</testVar></testFolder>   ", str(rfh2["testFolder"])))
            self.assertEqual(rfh2["mcdLength"], 28,
                             "mcdLength has incorrect value. Should be: %i But is: %s"
                             % (28, str(rfh2["mcdLength"])))
            self.assertEqual(rfh2["mcd"], b"<mcd><Msd>xmlnsc</Msd></mcd>",
                             "mcd has incorrect value. Should be: %s But is: %s"
                             % ("<mcd><Msd>xmlnsc</Msd></mcd>", str(rfh2["mcd"])))

        except Exception as e:
            self.fail(e)

    def test_parse_rfh2_with_incorrect_encoding(self):
        """Parse known correct message with its incorrect encoding.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message, 546)
        except mq.PYIFError as e:
            self.assertEqual(str(e), "PYIBMMQ Error: RFH2 - Buffer too short. Expected: 469827584 Buffer Length: 333",
                             "Exception Does not mathc expected. Expected: %s But is: %s"
                             % ("PYIBMMQ Error: RFH2 - Buffer too short. Expected: 469827584 Buffer Length: 333", str(e)))

    def test_add_folder(self):
        """Test the addition of a XML folder.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message)
            self.assertEqual(len(rfh2.get()), 14,
                             "Number of attributes incorrect.  Should be 12? But is %s"
                             % str(len(rfh2.get())))
            self.assertEqual(rfh2["_StrucId"], CMQC.MQRFH_STRUC_ID,
                             "StrucId has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQRFH_STRUC_ID, str(rfh2["_StrucId"])))
            self.assertEqual(rfh2["Version"], CMQC.MQRFH_VERSION_2,
                             "Version has incorrect value. Should be: %i But is: %s"
                             % (CMQC.MQRFH_VERSION_2, str(rfh2["Version"])))
            self.assertEqual(rfh2["StrucLength"], 284,
                             "StrucLength has incorrect value. Should be: %i But is: %s"
                             % (284, str(rfh2["StrucLength"])))
            self.assertEqual(rfh2["Encoding"], 273,
                             "Encoding has incorrect value. Should be: %i But is: %s"
                             % (273, str(rfh2["Encoding"])))
            self.assertEqual(rfh2["CodedCharSetId"], 1208,
                             "CodedCharSetId has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["CodedCharSetId"])))
            self.assertEqual(rfh2["Format"], CMQC.MQFMT_STRING,
                             "Format has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQFMT_NONE, str(rfh2["Format"])))
            self.assertEqual(rfh2["Flags"], 0,
                             "Flags has incorrect value. Should be: %i But is: %s"
                             % (0, str(rfh2["Flags"])))
            self.assertEqual(rfh2["NameValueCCSID"], 1208,
                             "NameValueCCSID has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["NameValueCCSID"])))
            self.assertEqual(rfh2["pscLength"], 152,
                             "pscLength has incorrect value. Should be: %i But is: %s"
                             % (152, str(rfh2["pscLength"])))
            self.assertEqual(rfh2["psc"],
                             b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ",
                             "psc has incorrect value. Should be: %s But is: %s"
                             % ("<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ", ">" + str(rfh2["psc"]) + "<"))
            self.assertEqual(rfh2["testFolderLength"], 56,
                             "testFolderLength has incorrect value. Should be: %i But is: %s"
                             % (56, str(rfh2["testFolderLength"])))
            self.assertEqual(rfh2["testFolder"], b"<testFolder><testVar>testValue</testVar></testFolder>   ",
                             "testFolder has incorrect value. Should be: %s But is: %s"
                             % ("<testFolder><testVar>testValue</testVar></testFolder>   ", str(rfh2["testFolder"])))
            self.assertEqual(rfh2["mcdLength"], 28,
                             "mcdLength has incorrect value. Should be: %i But is: %s"
                             % (28, str(rfh2["mcdLength"])))
            self.assertEqual(rfh2["mcd"], b"<mcd><Msd>xmlnsc</Msd></mcd>",
                             "mcd has incorrect value. Should be: %s But is: %s"
                             % ("<mcd><Msd>xmlnsc</Msd></mcd>", str(rfh2["mcd"])))

            rfh2.add_folder(b"<testFolder2><testVar>testValue</testVar></testFolder2>")

            self.assertEqual(len(rfh2.get()), 16,
                             "Number of attributes incorrect.  Should be 12? But is %s"
                             % str(len(rfh2.get())))
            self.assertEqual(rfh2["_StrucId"], CMQC.MQRFH_STRUC_ID,
                             "StrucId has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQRFH_STRUC_ID, str(rfh2["_StrucId"])))
            self.assertEqual(rfh2["Version"], CMQC.MQRFH_VERSION_2,
                             "Version has incorrect value. Should be: %i But is: %s"
                             % (CMQC.MQRFH_VERSION_2, str(rfh2["Version"])))
            self.assertEqual(rfh2["StrucLength"], 344,
                             "StrucLength has incorrect value. Should be: %i But is: %s"
                             % (344, str(rfh2["StrucLength"])))
            self.assertEqual(rfh2["Encoding"], 273,
                             "Encoding has incorrect value. Should be: %i But is: %s"
                             % (273, str(rfh2["Encoding"])))
            self.assertEqual(rfh2["CodedCharSetId"], 1208,
                             "CodedCharSetId has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["CodedCharSetId"])))
            self.assertEqual(rfh2["Format"], CMQC.MQFMT_STRING,
                             "Format has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQFMT_NONE, str(rfh2["Format"])))
            self.assertEqual(rfh2["Flags"], 0,
                             "Flags has incorrect value. Should be: %i But is: %s"
                             % (0, str(rfh2["Flags"])))
            self.assertEqual(rfh2["NameValueCCSID"], 1208,
                             "NameValueCCSID has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["NameValueCCSID"])))
            self.assertEqual(rfh2["pscLength"], 152,
                             "pscLength has incorrect value. Should be: %i But is: %s"
                             % (152, str(rfh2["pscLength"])))
            self.assertEqual(rfh2["psc"],
                             b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ",
                             "psc has incorrect value. Should be: %s But is: %s"
                             % ("<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ", ">" + str(rfh2["psc"]) + "<"))
            self.assertEqual(rfh2["testFolderLength"], 56,
                             "testFolderLength has incorrect value. Should be: %i But is: %s"
                             % (56, str(rfh2["testFolderLength"])))
            self.assertEqual(rfh2["testFolder"],
                             b"<testFolder><testVar>testValue</testVar></testFolder>   ",
                             "testFolder has incorrect value. Should be: %s But is: %s"
                             % ("<testFolder><testVar>testValue</testVar></testFolder>   ", str(rfh2["testFolder"])))
            self.assertEqual(rfh2["mcdLength"], 28,
                             "mcdLength has incorrect value. Should be: %i But is: %s"
                             % (28, str(rfh2["mcdLength"])))
            self.assertEqual(rfh2["mcd"], b"<mcd><Msd>xmlnsc</Msd></mcd>",
                             "mcd has incorrect value. Should be: %s But is: %s"
                             % ("<mcd><Msd>xmlnsc</Msd></mcd>", str(rfh2["mcd"])))
            self.assertEqual(rfh2["testFolder2Length"], 56,
                             "testFolderLength has incorrect value. Should be: %i But is: %s"
                             % (56, str(rfh2["testFolderLength"])))
            self.assertEqual(rfh2["testFolder2"], b"<testFolder2><testVar>testValue</testVar></testFolder2> ",
                             "testFolder2 has incorrect value. Should be: %s But is: %s"
                             % ("<testFolder2><testVar>testValue</testVar></testFolder2> ", str(rfh2["testFolder2"])))

        except Exception as e:
            self.fail(e)

    def test_rfh2_pack(self):
        """Test pack() by comparing output of pack() against original message.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message)
            self.assertEqual(len(rfh2.get()), 14,
                             "Number of attributes incorrect.  Should be 12? But is %s"
                             % str(len(rfh2.get())))
            self.assertEqual(rfh2["_StrucId"], CMQC.MQRFH_STRUC_ID,
                             "StrucId has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQRFH_STRUC_ID, str(rfh2["_StrucId"])))
            self.assertEqual(rfh2["Version"], CMQC.MQRFH_VERSION_2,
                             "Version has incorrect value. Should be: %i But is: %s"
                             % (CMQC.MQRFH_VERSION_2, str(rfh2["Version"])))
            self.assertEqual(rfh2["StrucLength"], 284,
                             "StrucLength has incorrect value. Should be: %i But is: %s"
                             % (284, str(rfh2["StrucLength"])))
            self.assertEqual(rfh2["Encoding"], 273,
                             "Encoding has incorrect value. Should be: %i But is: %s"
                             % (273, str(rfh2["Encoding"])))
            self.assertEqual(rfh2["CodedCharSetId"], 1208,
                             "CodedCharSetId has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["CodedCharSetId"])))
            self.assertEqual(rfh2["Format"], CMQC.MQFMT_STRING,
                             "Format has incorrect value. Should be: %s But is: %s"
                             % (CMQC.MQFMT_NONE, str(rfh2["Format"])))
            self.assertEqual(rfh2["Flags"], 0,
                             "Flags has incorrect value. Should be: %i But is: %s"
                             % (0, str(rfh2["Flags"])))
            self.assertEqual(rfh2["NameValueCCSID"], 1208,
                             "NameValueCCSID has incorrect value. Should be: %i But is: %s"
                             % (1208, str(rfh2["NameValueCCSID"])))
            self.assertEqual(rfh2["pscLength"], 152,
                             "pscLength has incorrect value. Should be: %i But is: %s"
                             % (152, str(rfh2["pscLength"])))
            self.assertEqual(rfh2["psc"],
                             b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ",
                             "psc has incorrect value. Should be: %s But is: %s"
                             % ("<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc> ", ">" + str(rfh2["psc"]) + "<"))
            self.assertEqual(rfh2["testFolderLength"], 56,
                             "testFolderLength has incorrect value. Should be: %i But is: %s"
                             % (56, str(rfh2["testFolderLength"])))
            self.assertEqual(rfh2["testFolder"],
                             b"<testFolder><testVar>testValue</testVar></testFolder>   ",
                             "testFolder has incorrect value. Should be: %s But is: %s"
                             % ("<testFolder><testVar>testValue</testVar></testFolder>   ", str(rfh2["testFolder"])))
            self.assertEqual(rfh2["mcdLength"], 28,
                             "mcdLength has incorrect value. Should be: %i But is: %s"
                             % (28, str(rfh2["mcdLength"])))
            self.assertEqual(rfh2["mcd"], b"<mcd><Msd>xmlnsc</Msd></mcd>",
                             "mcd has incorrect value. Should be: %s But is: %s"
                             % ("<mcd><Msd>xmlnsc</Msd></mcd>", str(rfh2["mcd"])))
            self.assertEqual(self.single_rfh2_message[0:rfh2["StrucLength"]], rfh2.pack(),
                             "result of RFH2.pack() not equal to original buffer used in unpack?")
        except Exception as e:
            self.fail(e)

    def test_rfh2_create(self):
        """Test the creation of a brand new RFH2. Compare the resulting byte array against identical known correct message.
        """

        rfh2 = mq.RFH2()
        new_rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message)
            new_rfh2["_StrucId"] = CMQC.MQRFH_STRUC_ID
            new_rfh2["Version"] = CMQC.MQRFH_VERSION_2
            new_rfh2["StrucLength"] = 188
            new_rfh2["Encoding"] = 273
            new_rfh2["CodedCharSetId"] = 1208
            new_rfh2["Format"] = CMQC.MQFMT_STRING
            new_rfh2["Flags"] = 0
            new_rfh2["NameValueCCSID"] = 1208
            new_rfh2.add_folder(b"<psc><Command>RegSub</Command><Topic>$topictree/topiccat/topic</Topic><QMgrName>DebugQM</QMgrName><QName>PUBOUT</QName><RegOpt>PersAsPub</RegOpt></psc>")
            new_rfh2.add_folder(b"<testFolder><testVar>testValue</testVar></testFolder>")
            new_rfh2.add_folder(b"<mcd><Msd>xmlnsc</Msd></mcd>")
            self.assertEqual(self.single_rfh2_message[0:rfh2["StrucLength"]], new_rfh2.pack(encoding=273), "New RFH2 Header does not match publishmessage?")
        except Exception as e:
            self.fail(e)

    def test_incorrect_strucid_exception(self):
        """Test exception that occurs if StrucId is incorrect.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message[116:])
        except mq.PYIFError as e:
            self.assertEqual(str(e),
                             "PYIBMMQ Error: RFH2 - _StrucId not MQRFH_STRUC_ID. Value: {0}".format(b'ame>'),
                             "_StrucId not = '%s'" % CMQC.MQRFH_STRUC_ID)

    def test_buffer_too_short_for_default_rfh2_exception(self):
        """Test exception occurs when buffer is too short for default RFH2.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message[0:32])
        except mq.PYIFError as e:
            self.assertEqual(str(e),
                             'PYIBMMQ Error: RFH2 - Buffer too short. Should be 36+ bytes instead of 32',
                             'Not Buffer to short exception?')

    def test_buffer_too_short_for_complete_rfh2_exception(self):
        """Test exception occurs when buffer is too short for complete RFH2.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message[0:188])
        except mq.PYIFError as e:
            self.assertEqual(str(e),
                             "PYIBMMQ Error: RFH2 - Buffer too short. Expected: 284 Buffer Length: 188",
                             "Not Buffer to short to parse complete RFH2 exception?")

    def test_folder_not_well_formed_exception_on_parse(self):
        """Test exception when parsing a message that contains not well formed XML folder.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.unpack(self.single_rfh2_message_not_well_formed)
        except mq.PYIFError as e:
            self.assertEqual(str(e).count("XML Folder not well formed"), 1)

    def test_folder_not_well_formed_exception_on_add(self):
        """Test exception when adding a not well formed XML folder.
        """

        rfh2 = mq.RFH2()
        try:
            rfh2.add_folder(b"<a><b>c</a>")
        except mq.PYIFError as e:
            # Don't depend on the actual XML library getting used (lxml or
            # minidom produce different error messages)
            self.assertTrue(
                str(e).startswith("PYIBMMQ Error: RFH2 - XML Folder not well formed. Exception:"),
                "Not 'XML Folder not well formed' exception (add_folder): %s" % (e, ))

    def test_encoding_on_pack_big_endian(self):
        """Test that pack() creates numeric fields with correct encoding. Big endian Test.
        """

        try:
            rfh2 = mq.RFH2()
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_FLOAT_S390)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_NORMAL)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_DECIMAL_NORMAL)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_FLOAT_IEEE_NORMAL)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_NORMAL + CMQC.MQENC_DECIMAL_NORMAL)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_NORMAL + CMQC.MQENC_FLOAT_IEEE_NORMAL)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_DECIMAL_NORMAL + CMQC.MQENC_FLOAT_IEEE_NORMAL)[4:8], b"\x00\x00\x00\x02")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_NORMAL + CMQC.MQENC_DECIMAL_NORMAL + CMQC.MQENC_FLOAT_IEEE_NORMAL)[4:8], b"\x00\x00\x00\x02")
        except Exception as e:
            self.fail(e)

    def test_encoding_on_pack_small_endian(self):
        """Test that pack() creates numeric fields with correct encoding. Small endian Test.
        """

        try:
            rfh2 = mq.RFH2()
            self.assertEqual(rfh2.pack()[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_NATIVE)[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_REVERSED)[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_DECIMAL_REVERSED)[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_FLOAT_IEEE_REVERSED)[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_REVERSED + CMQC.MQENC_DECIMAL_REVERSED)[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_INTEGER_REVERSED + CMQC.MQENC_FLOAT_IEEE_REVERSED)[4:8], b"\x02\x00\x00\x00")
            self.assertEqual(rfh2.pack(encoding=CMQC.MQENC_DECIMAL_REVERSED + CMQC.MQENC_FLOAT_IEEE_REVERSED)[4:8], b"\x02\x00\x00\x00")
        except Exception as e:
            self.fail(e)


if __name__ == '__main__':
    unittest.main()



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows how to put a message with a specific context value. It will
fail (2035) if the user does not have the authority to use the option.
"""
import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
message = 'Hello from Python!'
context_applid = 'MyApplId'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info)

od = mq.OD()
od.ObjectName = queue_name

queue = mq.Queue(qmgr)
queue.open(od, mq.CMQC.MQOO_OUTPUT | mq.CMQC.MQOO_SET_IDENTITY_CONTEXT)

md = mq.MD()
md.ApplIdentityData = context_applid

pmo = mq.PMO()
pmo.Options = mq.CMQC.MQPMO_SET_IDENTITY_CONTEXT

queue.put(message, md, pmo)

queue.close()
qmgr.disconnect()

print("Done.")



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows a simple and single immediate message retrieval. The message body
is the returned value from the get() function if it is successful. There is no waiting
for any message to arrive on the queue.
"""

import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
queue_name = 'DEV.QUEUE.1'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

queue = mq.Queue(qmgr, queue_name)
message = queue.get()
print("Message:", message)
queue.close()

qmgr.disconnect()

print("Done.")



"""MQSD: Subscription Descriptor"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class SD(MQOpts):
    """ Construct an MQSD Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQSD_STRUC_ID, '4s'],
                ['Version', CMQC.MQSD_VERSION_1, MQLONG_TYPE],
                ['Options', CMQC.MQSO_NON_DURABLE, MQLONG_TYPE],
                ['ObjectName', b'', '48s'],
                ['AlternateUserId', b'', '12s'],
                ['AlternateSecurityId', CMQC.MQSID_NONE, '40s'],
                ['SubExpiry', CMQC.MQEI_UNLIMITED, MQLONG_TYPE],

                # ObjectString
                ['ObjectStringVSPtr', 0, 'P'],
                ['ObjectStringVSOffset', (0), MQLONG_TYPE],
                ['ObjectStringVSBufSize', (0), MQLONG_TYPE],
                ['ObjectStringVSLength', (0), MQLONG_TYPE],
                ['ObjectStringVSCCSID', (0), MQLONG_TYPE],

                # Subname
                ['SubNameVSPtr', 0, 'P'],
                ['SubNameVSOffset', (0), MQLONG_TYPE],
                ['SubNameVSBufSize', (0), MQLONG_TYPE],
                ['SubNameVSLength', (0), MQLONG_TYPE],
                ['SubNameVSCCSID', (0), MQLONG_TYPE],

                # SubUserData
                ['SubUserDataVSPtr', 0, 'P'],
                ['SubUserDataVSOffset', (0), MQLONG_TYPE],
                ['SubUserDataVSBufSize', (0), MQLONG_TYPE],
                ['SubUserDataVSLength', (0), MQLONG_TYPE],
                ['SubUserDataVSCCSID', (0), MQLONG_TYPE],

                ['SubCorrelId', CMQC.MQCI_NONE, '24s'],
                ['PubPriority', CMQC.MQPRI_PRIORITY_AS_Q_DEF, MQLONG_TYPE],
                ['PubAccountingToken', CMQC.MQACT_NONE, '32s'],
                ['PubApplIdentityData', b'', '32s'],

                # SelectionString
                ['SelectionStringVSPtr', 0, 'P'],
                ['SelectionStringVSOffset', (0), MQLONG_TYPE],
                ['SelectionStringVSBufSize', (0), MQLONG_TYPE],
                ['SelectionStringVSLength', (0), MQLONG_TYPE],
                ['SelectionStringVSCCSID', (0), MQLONG_TYPE],

                ['SubLevel', 1, MQLONG_TYPE],

                # SelectionString
                ['ResObjectStringVSPtr', 0, 'P'],
                ['ResObjectStringVSOffset', (0), MQLONG_TYPE],
                ['ResObjectStringVSBufSize', (0), MQLONG_TYPE],
                ['ResObjectStringVSLength', (0), MQLONG_TYPE],
                ['ResObjectStringVSCCSID', (0), MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)



"""MQGMO: Get Message Options"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC, ibmmqc

class GMO(MQOpts):
    """ Construct an MQGMO Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQGMO_STRUC_ID, '4s'],
                ['Version', CMQC.MQGMO_VERSION_4, MQLONG_TYPE],
                ['Options', CMQC.MQGMO_NO_WAIT, MQLONG_TYPE],
                ['WaitInterval', 0, MQLONG_TYPE],
                ['Signal1', 0, MQLONG_TYPE],
                ['Signal2', 0, MQLONG_TYPE],
                ['ResolvedQName', b'', '48s'],
                ['MatchOptions', CMQC.MQMO_MATCH_MSG_ID + CMQC.MQMO_MATCH_CORREL_ID, MQLONG_TYPE],
                ['GroupStatus', CMQC.MQGS_NOT_IN_GROUP, 'b'],
                ['SegmentStatus', CMQC.MQSS_NOT_A_SEGMENT, 'b'],
                ['Segmentation', CMQC.MQSEG_INHIBITED, 'b'],
                ['_Reserved1', b' ', 'c'],
                ['MsgToken', b'', '16s'],
                ['ReturnedLength', CMQC.MQRL_UNDEFINED, MQLONG_TYPE], ]

        gmo_current_version = ibmmqc.__strucversions__.get("gmo", 1)
        if gmo_current_version >= CMQC.MQGMO_VERSION_4:
            opts += [
                ['_Reserved2', (0), MQLONG_TYPE],
                ['MsgHandle', (0), 'q']]

        super().__init__(tuple(opts), **kw)


# Backward compatibility
gmo = GMO  # pylint: disable=invalid-name



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and

"""
This example shows a way to connect to a remote queue manager using
the CSP class to provide credentials
"""
import ibmmq as mq

queue_manager = 'QM1'
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

csp = mq.CSP()
csp.CSPUserId = user
csp.CSPPassword = password

qmgr = mq.connect(queue_manager, channel, conn_info, csp=csp)
print("Connection succeeded.")
qmgr.disconnect()



"""MQDLH: Dead Letter Header"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class DLH(MQOpts):
    """ Construct an MQDLH Structure with default values as per MQI.
    The default values may be overridden by the optional keyword arguments 'kw'.
    """
    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQDLH_STRUC_ID, '4s'],
                ['Version', CMQC.MQDLH_VERSION_1, MQLONG_TYPE],
                ['Reason', CMQC.MQRC_NONE, MQLONG_TYPE],
                ['DestQName', b'', '48s'],
                ['DestQMgrName', b'', '48s'],
                ['Encoding', 0, MQLONG_TYPE],
                ['CodedCharSetId', CMQC.MQCCSI_UNDEFINED, MQLONG_TYPE],
                ['Format', CMQC.MQFMT_NONE, '8s'],
                ['PutApplType', 0, MQLONG_TYPE],
                ['PutApplName', b'', '28s'],
                ['PutDate', b'', '8s'],
                ['PutTime', b'', '8s']]

        super().__init__(tuple(opts), **kw)

    # Return the DLH as a formatted structure based on the message buffer. You
    # can then choose to also do a to_string() on the contents. This method
    # means you can pass the entire buffer, not just the DLH slice
    def get_header(self, buf):
        """Return the unpacked header from the full message buffer"""
        dlh = super().unpack(buf[:CMQC.MQDLH_LENGTH_1])
        return dlh



"""MQCBC: Callback Context"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from mqcommon import *
from mqopts import MQOpts
from ibmmq import CMQC

class CBC(MQOpts):
    """ Construct an MQCBC Structure with default values as per MQI. There
    are no real default values here, other than as needed to build the structure
    as the object is created and initialised by the qmgr rather than the application.
    """

    def __init__(self, **kw):
        opts = [['_StrucId', CMQC.MQCBC_STRUC_ID, '4s'],
                ['Version', CMQC.MQCBC_VERSION_2, MQLONG_TYPE],
                ['CallType', 0, MQLONG_TYPE],
                ['Hobj', 0, MQLONG_TYPE],
                ['CallbackArea', 0, 'P'],
                ['ConnectionArea', 0, 'P'],
                ['CompCode', 0, MQLONG_TYPE],
                ['Reason', 0, MQLONG_TYPE],
                ['State', 0, MQLONG_TYPE],
                ['DataLength', 0, MQLONG_TYPE],
                ['BufferLength', 0, MQLONG_TYPE],
                ['Flags', 0, MQLONG_TYPE],
                ['ReconnectDelay', 0, MQLONG_TYPE]]

        # Some padding is needed to match the C structure, even though it's not
        # listed in cmqc.h
        opts += [['_Reserved1', 0, MQLONG_TYPE]]

        super().__init__(tuple(opts), **kw)



"""Utility functions for handling different ways of opening queues"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from typing import Union
from ibmmq import ensure_strings_are_bytes, MD, OD, PMO, GMO

def common_q_args(*opts):
    """ Process args common to put/get/put1. Module Private.
    """
    ln = len(opts)
    if ln > 3:  # The 3rd parm might be passed in from a kwargs but it's not needed here
        raise TypeError('Too many args')
    m_desc = None
    pg_opts = None
    if ln > 0:
        m_desc = opts[0]
    if ln >= 2:
        pg_opts = opts[1]
    if m_desc is None:
        m_desc = MD()
    if not isinstance(m_desc, MD):
        raise TypeError("Message Descriptor must be an instance of MD")
    if pg_opts:
        if not isinstance(pg_opts, (PMO, GMO)):
            raise TypeError("Options must be an instance of PMO or GMO")

    return m_desc, pg_opts


# Backward compatibility
commonQArgs = common_q_args

# Some support functions for Queue ops.
def _make_q_desc(qdesc_or_string: Union[str, bytes, OD]) -> OD:
    """Maybe make MQOD from string."""
    if isinstance(qdesc_or_string, (str, bytes)):
        return OD(ObjectName=ensure_strings_are_bytes(qdesc_or_string))
    return ensure_strings_are_bytes(qdesc_or_string)



# More examples are at https://github.com/ibm-messaging/mq-dev-patterns
# and in code/examples in the source distribution.

"""
This example shows publishing a message to a topic
"""

from datetime import datetime
import ibmmq as mq

queue_manager = "QM1"
channel = 'DEV.APP.SVRCONN'
host = '127.0.0.1'
port = '1414'
conn_info = '%s(%s)' % (host, port)
user = 'app'
password = 'password'

topic_string = 'dev/'

now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
message = 'Hello from Python! Published at ' + now

qmgr = mq.connect(queue_manager, channel, conn_info, user, password)

topic = mq.Topic(qmgr, topic_string=topic_string)
topic.open(open_opts=mq.CMQC.MQOO_OUTPUT)
topic.pub(message)
print("Message published OK")
topic.close()

qmgr.disconnect()

print("Done.")



"""MQINQ/MQSET implementation"""

# Copyright (c) 2025 IBM Corporation and other Contributors. All Rights Reserved.
# Copyright (c) 2009-2024 Dariusz Suchojad. All Rights Reserved.

from ibmmq import CMQC, MQMIError, ibmmqc, mqcommon
import mqlog

# This object deals with the lengths of attributes that may be processed
# by the MQSET/MQINQ calls. Only a small set of the object attributes
# are supported by MQINQ (and even fewer for MQSET) so it's reasonable
# to list them all here. But any new attributes ought to be checked to see if
# they need to be added to the list.
mqAttrLength = {
    CMQC.MQCA_ALTERATION_DATE:       CMQC.MQ_DATE_LENGTH,
    CMQC.MQCA_ALTERATION_TIME:       CMQC.MQ_TIME_LENGTH,
    CMQC.MQCA_APPL_ID:               CMQC.MQ_PROCESS_APPL_ID_LENGTH,
    CMQC.MQCA_BACKOUT_REQ_Q_NAME:    CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_BASE_Q_NAME:           CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_CERT_LABEL:            CMQC.MQ_CERT_LABEL_LENGTH,
    CMQC.MQCA_QSG_CERT_LABEL:        CMQC.MQ_CERT_LABEL_LENGTH,
    CMQC.MQCA_CF_STRUC_NAME:         CMQC.MQ_CF_STRUC_NAME_LENGTH,
    CMQC.MQCA_CHANNEL_AUTO_DEF_EXIT: CMQC.MQ_EXIT_NAME_LENGTH,
    CMQC.MQCA_CHINIT_SERVICE_PARM:   CMQC.MQ_CHINIT_SERVICE_PARM_LENGTH,
    CMQC.MQCA_CLUSTER_NAME:          CMQC.MQ_CLUSTER_NAME_LENGTH,
    CMQC.MQCA_CLUSTER_NAMELIST:      CMQC.MQ_NAMELIST_NAME_LENGTH,
    CMQC.MQCA_CLUSTER_WORKLOAD_DATA: CMQC.MQ_EXIT_DATA_LENGTH,
    CMQC.MQCA_CLUSTER_WORKLOAD_EXIT: CMQC.MQ_EXIT_NAME_LENGTH,
    CMQC.MQCA_CLUS_CHL_NAME:         CMQC.MQ_OBJECT_NAME_LENGTH,
    CMQC.MQCA_COMMAND_INPUT_Q_NAME:  CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_COMM_INFO_NAME:        CMQC.MQ_OBJECT_NAME_LENGTH,
    CMQC.MQCA_CONN_AUTH:             CMQC.MQ_AUTH_INFO_NAME_LENGTH,
    CMQC.MQCA_CREATION_DATE:         CMQC.MQ_DATE_LENGTH,
    CMQC.MQCA_CREATION_TIME:         CMQC.MQ_TIME_LENGTH,
    CMQC.MQCA_CUSTOM:                CMQC.MQ_CUSTOM_LENGTH,
    CMQC.MQCA_DEAD_LETTER_Q_NAME:    CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_DEF_XMIT_Q_NAME:       CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_DNS_GROUP:             CMQC.MQ_DNS_GROUP_NAME_LENGTH,
    CMQC.MQCA_ENV_DATA:              CMQC.MQ_PROCESS_ENV_DATA_LENGTH,
    CMQC.MQCA_IGQ_USER_ID:           CMQC.MQ_USER_ID_LENGTH,
    CMQC.MQCA_INITIATION_Q_NAME:     CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_INSTALLATION_DESC:     CMQC.MQ_INSTALLATION_DESC_LENGTH,
    CMQC.MQCA_INSTALLATION_NAME:     CMQC.MQ_INSTALLATION_NAME_LENGTH,
    CMQC.MQCA_INSTALLATION_PATH:     CMQC.MQ_INSTALLATION_PATH_LENGTH,
    CMQC.MQCA_LU62_ARM_SUFFIX:       CMQC.MQ_ARM_SUFFIX_LENGTH,
    CMQC.MQCA_LU_GROUP_NAME:         CMQC.MQ_LU_NAME_LENGTH,
    CMQC.MQCA_LU_NAME:               CMQC.MQ_LU_NAME_LENGTH,
    CMQC.MQCA_NAMELIST_DESC:         CMQC.MQ_NAMELIST_DESC_LENGTH,
    CMQC.MQCA_NAMELIST_NAME:         CMQC.MQ_NAMELIST_NAME_LENGTH,
    CMQC.MQCA_NAMES:                 CMQC.MQ_OBJECT_NAME_LENGTH * 256,  # Maximum length to allocate
    CMQC.MQCA_PARENT:                CMQC.MQ_Q_MGR_NAME_LENGTH,
    CMQC.MQCA_PROCESS_DESC:          CMQC.MQ_PROCESS_DESC_LENGTH,
    CMQC.MQCA_PROCESS_NAME:          CMQC.MQ_PROCESS_NAME_LENGTH,
    CMQC.MQCA_Q_DESC:                CMQC.MQ_Q_DESC_LENGTH,
    CMQC.MQCA_Q_MGR_DESC:            CMQC.MQ_Q_MGR_DESC_LENGTH,
    CMQC.MQCA_Q_MGR_IDENTIFIER:      CMQC.MQ_Q_MGR_IDENTIFIER_LENGTH,
    CMQC.MQCA_Q_MGR_NAME:            CMQC.MQ_Q_MGR_NAME_LENGTH,
    CMQC.MQCA_Q_NAME:                CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_QSG_NAME:              CMQC.MQ_QSG_NAME_LENGTH,
    CMQC.MQCA_REMOTE_Q_MGR_NAME:     CMQC.MQ_Q_MGR_NAME_LENGTH,
    CMQC.MQCA_REMOTE_Q_NAME:         CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_REPOSITORY_NAME:       CMQC.MQ_Q_MGR_NAME_LENGTH,
    CMQC.MQCA_REPOSITORY_NAMELIST:   CMQC.MQ_NAMELIST_NAME_LENGTH,
    CMQC.MQCA_SSL_CRL_NAMELIST:      CMQC.MQ_NAMELIST_NAME_LENGTH,
    CMQC.MQCA_SSL_CRYPTO_HARDWARE:   CMQC.MQ_SSL_CRYPTO_HARDWARE_LENGTH,
    CMQC.MQCA_SSL_KEY_REPOSITORY:    CMQC.MQ_SSL_KEY_REPOSITORY_LENGTH,
    CMQC.MQCA_STORAGE_CLASS:         CMQC.MQ_STORAGE_CLASS_LENGTH,
    CMQC.MQCA_TCP_NAME:              CMQC.MQ_TCP_NAME_LENGTH,
    CMQC.MQCA_TRIGGER_DATA:          CMQC.MQ_TRIGGER_DATA_LENGTH,
    CMQC.MQCA_USER_DATA:             CMQC.MQ_PROCESS_USER_DATA_LENGTH,
    CMQC.MQCA_VERSION:               CMQC.MQ_VERSION_LENGTH,
    CMQC.MQCA_XMIT_Q_NAME:           CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_STREAM_QUEUE_NAME:     CMQC.MQ_Q_NAME_LENGTH,
    CMQC.MQCA_INITIAL_KEY:           CMQC.MQ_INITIAL_KEY_LENGTH,
    CMQC.MQCA_SSL_KEY_REPO_PASSWORD: CMQC.MQ_SSL_ENCRYP_KEY_REPO_PWD_LEN,
}

def get_attr_info(attrs):
    """Return info about the list of input selectors"""
    char_attr_length = 0
    char_attr_count = 0
    int_attr_count = 0

    for attr in attrs:
        try:
            # If it's a CHAR attribute, it ought to be in this map
            v = mqAttrLength[attr]
            char_attr_count += 1
            char_attr_length += v
        except KeyError as e:
            # Only integer values in this range are valid for MQINQ/MQSET.
            # Any other attributes require PCF command processing.
            if CMQC.MQIA_FIRST <= attr <= CMQC.MQIA_LAST:
                int_attr_count += 1
            else:
                # This might happen if a new char attribute is available for MQINQ but has
                # not been added to the above list. In which case, inform the package owners.
                raise MQMIError(CMQC.MQCC_FAILED, CMQC.MQRC_SELECTOR_ERROR) from e

    return {'intAttrCount': int_attr_count, 'charAttrCount': char_attr_count, 'charAttrLen': char_attr_length}

def get_attr_length(attr):
    """What is the max length of a char attributes.
    Return 0 if unknown.
    """
    v = 0
    try:
        v = mqAttrLength[attr]
    except KeyError:
        pass
    return v

def common_inq(hconn, hobj, selectors):
    """Call MQINQ - renamed from 'inq' to match the 'common_set' function"""

    mqlog.trace_entry("inq:common_inq")
    is_list = True
    single_attr = None

    # For compatibility with the original version that only
    # accepted a single attribute. Convert it to a list
    # and, at the end, return the one value.
    if not isinstance(selectors, list):
        single_attr = selectors
        is_list = False
        selectors = [selectors]

    attr_info = get_attr_info(selectors)

    # Create an empty list that will be used for the returned attributes
    int_attrs = []

    # Returns intAttrs (list), charAttr (byte string), MQCC, MQRC
    rv = ibmmqc.MQINQ(hconn, hobj, selectors,
                      int_attrs,
                      attr_info['intAttrCount'],
                      attr_info['charAttrCount'],
                      attr_info['charAttrLen'])
    if rv[-1]:
        mqlog.trace_exit("inq:common_inq", ep=1, rc=rv[-1])
        raise MQMIError(rv[-2], rv[-1])

    #
    int_attrs = rv[0]
    char_attrs = rv[1]

    attrs = {}
    j = 0
    char_offset = 0
    for s in selectors:
        if CMQC.MQIA_FIRST <= s <= CMQC.MQIA_LAST:
            attrs[s] = int_attrs[j]
            j += 1
        else:
            # If we had Namelist objects, more work would be needed to
            # deal with the lists of names returned from MQINQ
            char_length = get_attr_length(s)

            # The queue manager will not return the real key associated with this attribute,
            # but we want to make it printable.
            if s == CMQC.MQCA_INITIAL_KEY:
                v = b"********"
            else:
                v = char_attrs[char_offset:char_offset + char_length]
            char_offset += char_length

            # Turn into a string? Use default encoding?
            attrs[s] = v  # .decode().strip()

    # The backwards compatibility option - return the actual value instead of a map
    if not is_list:
        return attrs[single_attr]

    mqlog.trace_exit("inq:common_inq")
    return attrs

# For MQSET, the input is usually a dict of {selector:value}, similar to what's
# returned from MQINQ. But for backwards compatibility, we allow a single separate pair
# of parameters.
def common_set(hconn, hobj, *args):
    """Call MQSET - renamed from 'set' to avoid overloading language function"""
    mqlog.trace_entry("inq:common_set")

    temp = args[0][0]
    if not isinstance(temp, dict):
        # Convert an old-style k/v pair into a dict
        kv = {temp: args[0][1]}
    else:
        kv = temp

    selectors = []
    int_attrs = []
    char_attrs = b''

    # Build the list of IntAttrs and the CharAttr byte buffer from the input dict.
    # There's actually only a single char item that can be set for a queue (none for a qmgr)
    # but this handles it in a generic fashion.
    for k, v in kv.items():
        selectors.append(k)
        if CMQC.MQIA_FIRST <= k <= CMQC.MQIA_LAST:
            int_attrs.append(v)
        else:
            ll = get_attr_length(k)
            v = mqcommon.ensure_strings_are_bytes(v)
            # Turn the input string into a byte buffer of exactly the right length
            # Throw an immediate error if it's too long
            pad_len = ll - len(v)
            if pad_len < 0:
                mqlog.trace_exit("inq:common_set", ep=1)
                raise MQMIError(CMQC.MQCC_FAILED, CMQC.MQRC_CHAR_ATTR_LENGTH_ERROR)
            char_attrs += v + (pad_len * b'\0')

    rv = ibmmqc.MQSET(hconn, hobj, selectors, int_attrs, char_attrs)

    mqlog.trace_exit("inq:common_set")
    return rv



"""Propagate any OpenTelemetry context between the Python environment and MQ message properties. The application
is assumed to be already instrumented, and the OTel libraries available to be called. If the OTel libraries are
not accessible, then this module does not do anything.
"""

# Copyright (c) 2025, 2026 IBM Corporation and other Contributors. All Rights Reserved.

try:
    from opentelemetry import trace as oteltrace
    otel_tracer_enabled = True
except ImportError:
    otel_tracer_enabled = False
    raise

from os import environ
from threading import Lock

import mqlog
from mqcommon import *
from mqerrors import *
from ibmmq import CMQC, MessageHandle, Queue, OD, PD, IMPO, SMPO, RFH2
# from mqqmgr import *

class PropOptions:
    """Keep track of the options used for a GET so we can preserve them across the Before/After calls.
    This class is only used in this module; more effort than it's worth to have the full setter/getter/init routines
    """
    prop_ctl = -1  # The PROPCTL attribute on the queue, or -1 if unknown
    gmo = 0  # Currently-active GMO Options value so we can reset
    managed_ho = None  # To link a managed queue with its corresponding MQSUB hObj


object_options = {}
object_handle = {}

object_options_lock = Lock()
object_handle_lock = Lock()

# The property names that we work with
traceparent = "traceparent"
tracestate = "tracestate"

# Use this as a bitmap filter to pull out relevant value from GMO.
# The AS_Q_DEF value is 0 so would not contribute.
get_props_options = CMQC.MQGMO_PROPERTIES_FORCE_MQRFH2 |\
    CMQC.MQGMO_PROPERTIES_IN_HANDLE |\
    CMQC.MQGMO_NO_PROPERTIES |\
    CMQC.MQGMO_PROPERTIES_COMPATIBILITY

# Options in an MQOPEN that mean we might do MQGET
# Do not include BROWSE variants
open_get_options = CMQC.MQOO_INPUT_AS_Q_DEF |\
    CMQC.MQOO_INPUT_SHARED |\
    CMQC.MQOO_INPUT_EXCLUSIVE


# This function can be useful to create a unique key related to the hConn and hObj values
# where the object name is not sufficiently unique
def _make_key(hc, ho) -> str:
    suffix = "*"
    if ho is not None:
        if isinstance(ho, int):
            if ho not in (CMQC.MQHO_NONE, CMQC.MQHO_UNUSABLE_HOBJ):
                suffix = str(ho)
        else:
            suffix = str(ho.get_handle())

    if isinstance(hc, int):
        prefix = str(hc)
    else:
        prefix = str(hc.get_handle())

    k = prefix + "/" + suffix
    mqlog.debug(f"make_key: {k}")
    return k

def _is_usable_handle(mh):
    """Is the message handle valid? We are given the actual integer value, not a MessageHandle object"""
    rc = False
    if mh not in (CMQC.MQHM_NONE, CMQC.MQHM_UNUSABLE_HMSG):
        rc = True

    mqlog.debug(f"is_usable_handle: {mh} {rc}")
    return rc

# Do we have a MsgHandle for this hConn? If not, create a new one
def _get_msg_handle(hc, ho) -> MessageHandle:
    key = _make_key(hc, ho)
    with object_handle_lock:
        if object_handle.get(key) is None:
            try:
                mh = MessageHandle(qmgr=hc)
                mqlog.debug(f"Adding handle for key {key}")
                object_handle[key] = mh
            except MQMIError as e:
                mqlog.error(e)
                raise e

    o = object_handle.get(key)

    return o

# Is the GMO/PMO MsgHandle one that we allocated?
def _compare_msg_handle(hc, ho, mh_value) -> bool:
    rc = False
    key = _make_key(hc, ho)
    with object_handle_lock:
        oh = object_handle.get(key)
        if oh:
            mh_local = oh
            if mh_local.get_handle() == mh_value:
                rc = True

    return rc

def _props_contain(mh: MessageHandle, prop: str) -> bool:
    """ Is there a property of the given name?"""
    rc = False

    pd = PD()
    impo = IMPO()
    impo.Options = CMQC.MQIMPO_CONVERT_VALUE | CMQC.MQIMPO_INQ_FIRST

    # Don't care about the actual value of the property, just that it exists.
    try:
        mh.get(prop, impo=impo, pd=pd)
        rc = True
    except MQMIError:
        pass

    return rc


# Extract a substring from the RFH2 properties in a given folder
def _extract_rfh2_prop_val(props: bytes, prop: str) -> str:
    props_str = str(props, "utf-8")
    prop_xml = "<" + prop + ">"
    val = ""

    idx = props_str.find(prop_xml)
    if idx != -1:
        start = props_str[idx + len(prop_xml):]
        # Where does the next tag begin
        end = start.find("<")
        if end != -1:
            val = start[0:end]

    mqlog.debug(f"Searched for {prop} in RFH2 \"{props}\". Found: \"{val}\"")
    return val


def otel_disc(hc):
    """Get rid of entries in the hconn/hobj maps when the application calls MQDISC
     """

    mqlog.trace_entry("otel:disc")
    # Both the maps are keyed by a string which begins with
    # the hConn value. As this is MQDISC, we don't care about
    # any specific hObj
    prefix = str(hc.get_handle()) + "/"
    mqlog.debug(f"otel:disc prefix={prefix}")
    with object_handle_lock:
        for k in object_handle.copy():  # Use copy to avoid changing dict underneath the iteration
            if k.startswith(prefix):
                mh = object_handle[k]
                try:
                    mh.dlt()
                except MQMIError as e:
                    mqlog.error(f"MessageHandle delete error: {e}")
                    # pass
                mqlog.debug(f"Deleting handle for key {k}")
                del object_handle[k]

    # And delete information about any OPENed object too
    with object_options_lock:
        for k in object_options.copy():
            if k.startswith(prefix):
                mho = object_options[k].managed_ho
                if mho:
                    otel_close_nolock(mho)
                mqlog.debug(f"Deleting options for key {k}")
                del object_options[k]
    mqlog.trace_exit("otel:disc")

def otel_open(ho, od, open_options, managed_ho):
    """When a queue is opened for INPUT, then it will help to
    know the PROPCTL setting so we know if we can add a MsgHandle or to expect
    an RFH2 response. If the MQINQ fails, that's OK - we'll just ignore the error
    but might not be able to get any property/RFH from an inbound message

    Note that we can't (and don't need to) do the same for an MQPUT1 because the
    information we are trying to discover is only useful on MQGET/CallBack.
     """

    prop_ctl = -1

    mqlog.trace_entry("otel:open")
    mqlog.trace(f"otel:open: hobjReal={ho.get_handle()} open_options={open_options} mho={managed_ho}")

    # Do the MQINQ and stash the information
    # Only care if there's an INPUT (MQGET) option. We do the MQINQ on every relevant MQOPEN
    # because it might change between an MQCLOSE and a subsequent MQOPEN. The MQCLOSE
    # will, in any case, have discarded the entry from this map.
    # If the user opened the queue with MQOO_INQUIRE, then we can reuse the object handle.
    # Otherwise we have to do our own open/inq/close.
    # If the MQOPEN is for a topic (via MQSUB) AND there's a managed object created for the
    # target queue, then that will always have the MQOO_INQUIRE setup for us.
    if ((od and od.ObjectType == CMQC.MQOT_Q) and (open_options & open_get_options) != 0) or (managed_ho is not None):
        hc = ho.get_queue_manager()
        key = _make_key(hc, ho)
        prop_ctl = 0
        selectors = [CMQC.MQIA_PROPERTY_CONTROL]
        if managed_ho is None and (open_options & CMQC.MQOO_INQUIRE) != 0:
            mqlog.debug("otel:open: Reusing existing hObj")
            try:
                values = ho.inquire(selectors)
                mqlog.debug(f"Inq Responses: {values}")
                prop_ctl = values[selectors[0]]
            except MQMIError as e:
                mqlog.error(f"otel:open: Inq err {e}")
                prop_ctl = -1

        elif managed_ho is not None:
            mqlog.debug("open: Using managed hObj")
            try:
                values = managed_ho.inquire(selectors)
                mqlog.debug(f"Inq Responses: {values}")
                prop_ctl = values[selectors[0]]
            except MQMIError as e:
                mqlog.error(f"otel:open: Inq err {e}")
                prop_ctl = -1

            # And add this to the map so it can be referenced during MQGETs
            managed_key = _make_key(hc, managed_ho)
            options = PropOptions()
            options.prop_ctl = prop_ctl
            # replace any existing value for this object handle
            with object_options_lock:
                object_options[managed_key] = options

        else:
            inq_od = OD()
            inq_od.ObjectName = od.ObjectName
            inq_od.ObjectQMgrName = od.ObjectQMgrName
            inq_od.ObjectType = CMQC.MQOT_Q
            inq_open_options = CMQC.MQOO_INQUIRE

            mqlog.debug("otel:open: pre-Reopen")
            # This gets a little recursive as this Open will end up calling back into this function. But
            # as it's only doing MQOO_INQUIRE, then we don't nest any further
            try:
                inq_ho = Queue(hc, inq_od, inq_open_options)
                try:
                    values = inq_ho.inquire(selectors)
                    mqlog.debug(f"Inq Responses: {values}")
                    prop_ctl = values[selectors[0]]
                except MQMIError as e:
                    mqlog.error(f"otel:open: Inq err {e}")
                    prop_ctl = -1

                try:
                    inq_ho.close()  # Ignore any error
                except MQMIError:
                    pass

            except MQMIError as e:
                mqlog.error(f"otel:open: Reopen err {e}")
                prop_ctl = -1

        # Create an object to hold the discovered value
        options = PropOptions()
        options.prop_ctl = prop_ctl
        options.managed_ho = managed_ho
        # replace any existing value for this object handle
        with object_options_lock:
            object_options[key] = options

    else:
        mqlog.trace("open: not doing Inquire")

    mqlog.trace_exit("otel:open")

def otel_close(ho):
    """Called during the MQCLOSE"""
    mqlog.trace_entry("otel:close")
    key = _make_key(ho.get_queue_manager(), ho)
    with object_options_lock:
        if object_options.get(key):
            mho = object_options[key].managed_ho
            if mho:
                otel_close_nolock(mho)
            mqlog.debug(f"Deleting options for key {key}")
            del object_options[key]
    mqlog.trace_exit("otel:close")

def otel_close_nolock(ho):
    """Called during the MQCLOSE of a subscription to cleanup any managed
    queues. The object lock is already held when this is called
    """
    mqlog.trace_entry("otel:close_nolock")

    key = _make_key(ho.get_queue_manager(), ho)
    if object_options.get(key):
        mqlog.debug(f"Deleting options for key {key}")
        del object_options[key]
    mqlog.trace_exit("otel:close_nolock")

def otel_put_trace_before(hc, md, pmo, buffer):
    """Insert any span-provided properties from the environment"""

    mh = None
    mho = None

    mqlog.trace_entry("otel:put_trace_before")

    skip_parent = False
    skip_state = False

    # Is the app already using a MsgHandle for its PUT? If so, we
    # can piggy-back on that. If not, then we need to use our
    # own handle. That handle can be reused for all PUTs/GETs on this
    # hConn. This works, even when the app is primarily using an RFH2 for
    # its own properties - the RFH2 and the Handle contents are merged.

    # If there was an app-provided handle, then have they set
    # either of the key properties? If so, then we will
    # leave them alone as we are not trying to create a new span in this layer.
    if _is_usable_handle(pmo.NewMsgHandle):
        mh = pmo.NewMsgHandle
        if _props_contain(mh, traceparent):
            skip_parent = True
        if _props_contain(mh, tracestate):
            skip_state = True

    elif _is_usable_handle(pmo.OriginalMsgHandle):
        mho = pmo.OriginalMsgHandle
        if _props_contain(mho, traceparent):
            skip_parent = True

        if _props_contain(mho, tracestate):
            skip_state = True

    else:
        # The PMO uses the real integer value of the msg handle
        mh = _get_msg_handle(hc, None).msg_handle
        pmo.OriginalMsgHandle = mh

    # Make sure we've got one of the handles set
    if _is_usable_handle(mho) and not _is_usable_handle(mh):
        mh = mho

    temp_msg_handle = MessageHandle(qmgr=hc, dup_handle=mh)

    # The message MIGHT have been constructed with an explicit RFH2
    # header. Unlikely, but possible as we tend to prefer properties. If so, then we extract the properties
    # from that header (assuming there's only a single structure, and it's not
    # chained). Then very simply look for the property names in there as strings. These tests would
    # incorrectly succeed if someone had put "traceparent" into a non-"usr" folder but that would be
    # very unexpected.
    if md.Format == CMQC.MQFMT_RF_HEADER_2:
        rfh2 = RFH2()
        rfh2.unpack(buffer)
        try:
            folders = rfh2.get_folders()
            for folder in folders:
                props = rfh2[folder]
                if "<" + traceparent + ">" in props:
                    skip_parent = True
                if "<" + tracestate + ">" in props:
                    skip_state = True
        except KeyError:
            pass

    # We're now ready to extract the context information and set the MQ message property
    # We are not going to try to propagate baggage via another property
    span = oteltrace.get_current_span()
    mqlog.debug(f"Span: {span}")

    if span != oteltrace.INVALID_SPAN:
        span_context = span.get_span_context()
        mqlog.debug(f"Span/Context: {span_context}")

        if span_context.is_valid:
            smpo = SMPO()
            pd = PD()

            mqlog.debug("About to extract context from an active span")
            if not skip_parent:
                trace_id = span_context.trace_id
                span_id = span_context.span_id
                trace_flags = span_context.trace_flags
                trace_flags_string = "01"
                if trace_flags != 1:
                    trace_flags_string = "00"

                # This is the W3C-defined format for the trace property
                value = "00" + "-" + \
                    oteltrace.span.format_trace_id(trace_id) + "-" + \
                    oteltrace.span.format_span_id(span_id) + "-" + \
                    trace_flags_string

                mqlog.debug(f"Setting {traceparent} to {value}")

                try:
                    temp_msg_handle.properties.set(traceparent, value, smpo=smpo, pd=pd)
                except MQMIError:
                    # Fail silently?
                    pass

            if not skip_state:
                # Need to convert any TraceState map to a single serialised string
                ts = span_context.trace_state
                value = ts.to_header()
                if value is not None:  # and value != "":
                    mqlog.debug(f"Setting {tracestate} to \"{value}\"")
                    try:
                        temp_msg_handle.properties.set(tracestate, value, smpo=smpo, pd=pd)
                    except MQMIError:
                        # Fail silently?
                        pass
    else:
        mqlog.debug("No valid active span")

    mqlog.trace_exit("otel:put_trace_before")


def otel_put_trace_after(hc, pmo):
    """
    If we added our own MsgHandle to the PMO, then remove it
    before returning to the application. We don't need to delete
    the handle as it can be reused for subsequent PUTs on this hConn
    """
    mqlog.trace_entry("otel:put_trace_after")

    mh = pmo.OriginalMsgHandle
    if _compare_msg_handle(hc, None, mh):
        mqlog.debug("Replacing handle with default")
        pmo.OriginalMsgHandle = 0

    mqlog.trace_exit("otel:put_trace_after")

def otel_get_trace_before(hc, ho, gmo, asynchronous):
    """Decide whether or not to use a message handle when retrieving a message
    Called during MQGET and MQCB (that sets up the callback)
    """
    prop_ctl = 0

    mqlog.trace_entry("otel:get_trace_before")

    # Option combinations:
    # MQGMO_NO_PROPERTIES: Always add our own handle
    # MQGMO_PROPERTIES_IN_HANDLE: Use it
    # MQGMO_PROPERTIES_COMPAT/FORCE_RFH2: Any returned properties will be in RFH2
    # MQGMO_PROPERTIES_AS_Q_DEF:
    #      PROPCTL: NONE: same as GMO_NO_PROPERTIES
    #               ALL/COMPATV6COMPAT: Any returned properties will be either in RFH2 or Handle if supplied
    #               FORCE: Any returned properties will be in RFH2
    prop_get_options = gmo.Options & get_props_options
    mqlog.debug(f"propGetOptions: 0x{prop_get_options:08x}")

    if _is_usable_handle(gmo.MsgHandle):
        mqlog.debug("Using app-supplied msg handle")
    else:
        key = _make_key(hc, ho)
        prop_ctl = -1
        with object_options_lock:
            opts = object_options.get(key)
            if opts is not None:
                prop_ctl = opts.prop_ctl
                # Stash the GMO options so they can be restored afterwards
                opts.gmo = gmo.Options
                object_options[key] = opts

        # If we know that the app or queue is configured for not returning any properties, then we will override that into our handle
        if (prop_get_options == CMQC.MQGMO_NO_PROPERTIES) or (prop_get_options == CMQC.MQGMO_PROPERTIES_AS_Q_DEF and prop_ctl == CMQC.MQPROP_NONE):
            gmo.Options &= ~CMQC.MQGMO_NO_PROPERTIES
            gmo.Options |= CMQC.MQGMO_PROPERTIES_IN_HANDLE
            tmp_ho = ho
            if not asynchronous:
                tmp_ho = None

            gmo.MsgHandle = _get_msg_handle(hc, tmp_ho).get_handle()
            mqlog.debug(f"Using mqotel msg handle. get_props_options=0x{prop_get_options:08x} prop_ctl={prop_ctl}")
        else:
            # Hopefully they will have set something suitable on the PROPCTL attribute
            # or are asking specifically for an RFH2-style response
            mqlog.debug(f"Not setting a message handle. prop_get_options=0x{prop_get_options:08x}")

    mqlog.trace_exit("otel:get_trace_before")

def _int_from_hex(s: str, default: int) -> int:
    try:
        return int(s, 16)
    except ValueError:
        return default

def otel_get_trace_after(ho, gmo, md, otel_options, buffer, asynchronous):
    """ Extract the properties from the message, either with the properties API
    or from the RFH2. Construct an object with the span information.
    We do not try to extract/propagate any baggage-related fields.
    """
    mqlog.trace_entry("otel:get_trace_after")

    traceparent_val = ""
    tracestate_val = ""

    if buffer is None:
        mqlog.trace_exit("otel:get_trace_after", ep=1)
        return 0

    if otel_options:
        remove_rfh2 = otel_options.remove_rfh2
    else:
        remove_rfh2 = False
    mqlog.debug(f"OTelOptions={otel_options} removeRFH2={remove_rfh2}")

    removed = 0
    mh = gmo.MsgHandle
    if _is_usable_handle(mh):
        temp_msg_handle = MessageHandle(qmgr=hc, dup_handle=mh)

        pd = PD()
        impo = IMPO()
        impo.Options = CMQC.MQIMPO_CONVERT_VALUE | CMQC.MQIMPO_INQ_FIRST

        try:
            val = temp_msg_handle.inq(impo, pd, traceparent)
            mqlog.debug(f"Found traceparent property: {val}")
            traceparent_val = val
        except MQMIError as e:
            if e.reason != CMQC.MQRC_PROPERTY_NOT_AVAILABLE:
                # Should not happen
                mqlog.error(e)

        try:
            val = temp_msg_handle.inq(impo, pd, tracestate)
            mqlog.debug(f"Found tracestate property: {val}")
            tracestate_val = val
        except MQMIError as e:
            if e.reason != CMQC.MQRC_PROPERTY_NOT_AVAILABLE:
                # Should not happen
                mqlog.error(e)

        # If we added our own handle in the GMO, then reset
        # but don't do it for async callbacks.
        tmp_ho = ho
        hc = ho.get_queue_manager()
        if not asynchronous:
            tmp_ho = None

        if not asynchronous and _compare_msg_handle(hc, tmp_ho, mh):
            gmo.MsgHandle = 0
            key = _make_key(hc, tmp_ho)
            with object_options_lock:
                opts = object_options.get(key)
                if opts is not None:
                    gmo.Options = opts.gmo
                else:
                    gmo.Options &= ~CMQC.MQGMO_PROPERTIES_IN_HANDLE

            mqlog.debug(f"Removing our handle: hObj={tmp_ho}")

        # Should we also remove the properties?
        # Probably not worth it, as any app dealing with
        # properties ought to be able to handle unexpected props.

    elif md.Format == CMQC.MQFMT_RF_HEADER_2:
        rfh2 = RFH2()
        rfh2.unpack(buffer)
        try:
            folders = rfh2.get_folders()
            for folder in folders:
                props = rfh2[folder]
                traceparent_val = _extract_rfh2_prop_val(props, traceparent)
                tracestate_val = _extract_rfh2_prop_val(props, tracestate)
                if traceparent_val != "" or tracestate_val != "":
                    break
        except KeyError:
            pass

        # If the only properties in the RFH2 are the OTEL ones, then perhaps
        # the application cannot process the message. But we don't know for sure,
        # and maybe the properties are useful for higher-level span generation.
        # So we should have an option to forcibly remove the RFH2. Other bindings
        # have extended the GMO to give an extra option. That's hard to do
        # with the python way of handling MQI structures, so we do it
        # via an extra keyword parameter on get() and cb().
        #    otel_opts = {'remove_rfh2': True}
        if remove_rfh2:
            md.Format = rfh2.Format
            md.CodedCharSetId = rfh2.CodedCharSetId
            md.Encoding = rfh2.Encoding
            removed = rfh2.StrucLength

    # We now should have the relevant message properties to pass upwards
    trace_id = 0
    span_id = 0
    trace_flags = None
    trace_state = None

    span_context = None
    current_span = oteltrace.get_current_span()
    mqlog.debug(f"Span: {current_span}")

    if current_span != oteltrace.INVALID_SPAN:
        span_context = current_span.get_span_context()
        mqlog.debug(f"Span/Context: {span_context}")

    if span_context is not None and span_context.is_valid:

        have_new_context = False

        if traceparent_val != "":
            # Split the inbound traceparent value into its components to allow
            # construction of a new context
            elem = traceparent_val.split("-")
            if len(elem) == 4:
                # elem[0] = 0 (version indicator. Always 0 for now)

                trace_id = _int_from_hex(elem[1], oteltrace.INVALID_TRACE_ID)
                span_id = _int_from_hex(elem[2], oteltrace.INVALID_SPAN_ID)
                # Final element can only be 00 or 01 (for now)
                if elem[3] == "00":
                    trace_flags = oteltrace.TraceFlags.DEFAULT
                else:
                    trace_flags = oteltrace.TraceFlags.SAMPLED

                have_new_context = True

        if tracestate_val != "":
            # Build a TraceState structure by parsing the string
            trace_state = oteltrace.TraceState.from_header([tracestate_val])
            if trace_state is not None:
                have_new_context = True

        # If there is a current span, and we have at least one of the
        # parent/state properties, then create a link referencing these values
        if have_new_context:
            msg_context = oteltrace.SpanContext(trace_id=trace_id, span_id=span_id, is_remote=False, trace_flags=trace_flags, trace_state=trace_state)

            # mqlog.debug(f"Created new context: {msg_context}")
            if msg_context is not None:
                current_span.add_link(msg_context)
                mqlog.debug("Added link to current span")
                # mqlog.debug(f"Updated span: {currentSpan}")
            else:
                mqlog.debug("Unable to create a usable msg_context")

        else:
            mqlog.debug("No context properties found")

    else:
        # If there is no current active span, then we are not going to
        # try to create a new one, as we would have no way of knowing when it
        # ends. The properties are (probably) still available to the application if
        # it wants to work with them itself.
        mqlog.debug("No current active span to update")

    mqlog.debug(f"removed:{removed}")
    mqlog.trace_exit("otel:get_trace_after")
    return removed

def init():
    """Any initialisation operations needed for the OTel interface"""
    mqlog.trace_entry("otel:init")

    # Set the function pointers to invoke the code in here
    OTelFunctions.disc = otel_disc
    OTelFunctions.open = otel_open
    OTelFunctions.close = otel_close
    OTelFunctions.put_trace_before = otel_put_trace_before
    OTelFunctions.put_trace_after = otel_put_trace_after
    OTelFunctions.get_trace_before = otel_get_trace_before
    OTelFunctions.get_trace_after = otel_get_trace_after

    # Set this so any underlying equivalent code in the C library (my API Exit) will not try to do its own thing
    environ["AMQ_OTEL_INSTRUMENTED"] = "true"

    mqlog.trace_exit("otel:init")


if environ.get("MQIPY_NOOTEL", None) is None:
    init()



""" Tests for making sure modules generated automatically by h2py have no
unwanted duplicate entries.
"""

# stdlib
import inspect
import unittest
from struct import calcsize

from ibmmq import CMQC, CMQXC, CMQCFC

ignore_dups = []

class Testh2py(unittest.TestCase):
    """ Test that modules generated automatically by h2py have no unwanted duplicate entries.
    """

    def test_h2py(self):
        """ Test that modules generated automatically by h2py have no unwanted duplicate entries.
        """

        errors = []
        test_passed = False

        for mod in CMQC, CMQXC, CMQCFC:
            mq_attrs = []
            mq_attrs_count = {}

            source_lines = inspect.getsourcelines(mod)
            if mod == CMQC:
                source_lines[0].append('MQ_DUMMY_1 = 1')
                source_lines[0].append('MQ_DUMMY_1 = 1')
                source_lines[0].append('if calcsize("P") == 8:')
                source_lines[0].append('    MQ_DUMMY_2 = 1')
                source_lines[0].append('else:')
                source_lines[0].append('    MQ_DUMMY_2 = 1')
            process = True
            in_if = False
            indent = 0
            for line in source_lines[0]:
                len_line = len(line)
                len_line_lstrip = len(line.lstrip(' '))
                line = line.strip()

                if line:
                    if "".join(line.split()) == ''.join('if calcsize("P") == 8:'.split()):
                        in_if = True
                        process = calcsize("P") == 8
                    elif "".join(line.split()) == 'else:' and in_if:
                        process = not process
                    else:
                        if in_if and not indent:
                            indent = len_line - len_line_lstrip
                        elif (len_line - len_line_lstrip) < indent and line[0] != '#':
                            in_if = False
                            process = True
                            indent = 0

                        if line.startswith("MQ") and process:
                            line_split = line.split()
                            mq_attr = line_split[0].split(",")[0]
                            if len(line_split) > 1 and line_split[1] == '=':
                                mq_attrs.append(mq_attr)

            for mq_attr in mq_attrs:
                if mq_attr not in mq_attrs_count:
                    mq_attrs_count[mq_attr] = 1
                else:
                    mq_attrs_count[mq_attr] += 1

            for mq_attr, count in mq_attrs_count.items():
                if count == 1 or (count == 2 and mq_attr in ignore_dups):
                    continue
                if mq_attr.startswith('MQ_DUMMY'):
                    test_passed = True
                else:
                    item = "%s is defined %d times in %s" % (mq_attr, count, mod)
                    errors.append(item)

        if errors:
            msg = "\n" + "\n".join(errors)
            self.fail(msg)

        self.assertTrue(test_passed, 'Test does not found DUMMY duplicate entries!')


if __name__ == "__main__":
    unittest.main()
